#!/usr/bin/env python

"""
High-level classes.

Copyright (C) 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.filesystem import *
from iixr.merging import *
from os import listdir, mkdir    # index and partition discovery
from os.path import exists

try:
    set
except NameError:
    from sets import Set as set

# Constants.

TERM_INTERVAL     = 100
DOCUMENT_INTERVAL = 100
FIELD_INTERVAL    = 100
FLUSH_INTERVAL    = 10000

# High-level classes.

class Document:

    "A container of document information."

    def __init__(self, docnum):
        self.docnum = docnum
        self.fields = []
        self.terms = {}

    def add_position(self, term, position):

        """
        Add a position entry for the given 'term', indicating the given
        'position'.
        """

        self.terms.setdefault(term, []).append(position)

    def add_field(self, identifier, value):

        "Add a field having the given 'identifier' and 'value'."

        self.fields.append((identifier, unicode(value))) # convert to string

    def set_fields(self, fields):

        """
        Set the document's 'fields': a list of tuples each containing an integer
        identifier and a string value.
        """

        self.fields = fields

class IndexWriter:

    """
    Building term information and writing it to the term and field dictionaries.
    """

    def __init__(self, pathname, interval, doc_interval, flush_interval):
        self.pathname = pathname
        self.interval = interval
        self.doc_interval = doc_interval
        self.flush_interval = flush_interval

        self.dict_partition = 0
        self.field_dict_partition = 0

        self.terms = {}
        self.docs = {}

        self.doc_counter = 0

    def add_document(self, doc):

        """
        Add the given document 'doc', updating the document counter and flushing
        terms and fields if appropriate.
        """

        for term, positions in doc.terms.items():
            self.terms.setdefault(term, {})[doc.docnum] = positions

        self.docs[doc.docnum] = doc.fields

        self.doc_counter += 1
        if self.flush_interval and self.doc_counter >= self.flush_interval:
            self.flush_terms()
            self.flush_fields()
            self.doc_counter = 0

    def get_term_writer(self):

        "Return a term dictionary writer for the current partition."

        return get_term_writer(self.pathname, self.dict_partition, self.interval, self.doc_interval)

    def get_field_writer(self):

        "Return a field dictionary writer for the current partition."

        return get_field_writer(self.pathname, self.field_dict_partition, self.interval)

    def flush_terms(self):

        "Flush terms into the current term dictionary partition."

        # Get the terms in order.

        all_terms = self.terms
        terms = all_terms.keys()
        terms.sort()

        dict_writer = self.get_term_writer()

        for term in terms:
            doc_positions = all_terms[term].items()
            dict_writer.write_term_positions(term, doc_positions)

        dict_writer.close()

        self.terms = {}
        self.dict_partition += 1

    def flush_fields(self):

        "Flush fields into the current term dictionary partition."

        # Get the documents in order.

        docs = self.docs.items()
        docs.sort()

        field_dict_writer = self.get_field_writer()

        for docnum, fields in docs:
            field_dict_writer.write_fields(docnum, fields)

        field_dict_writer.close()

        self.docs = {}
        self.field_dict_partition += 1

    def close(self):
        if self.terms:
            self.flush_terms()
        if self.docs:
            self.flush_fields()

class IndexReader:

    "Accessing the term and field dictionaries."

    def __init__(self, pathname):
        self.dict_reader = get_term_reader(pathname, "merged")
        self.field_dict_reader = get_field_reader(pathname, "merged")

    def find_terms(self, term):
        return self.dict_reader.find_terms(term)

    def find_positions(self, term):
        return self.dict_reader.find_positions(term)

    def find_common_positions(self, term):
        return self.dict_reader.find_common_positions(term)

    def get_frequency(self, term):
        return self.dict_reader.get_frequency(term)

    def get_document_frequency(self, term):
        return self.dict_reader.get_document_frequency(term)

    def get_fields(self, docnum):
        return self.field_dict_reader.get_fields(docnum)

    def close(self):
        self.dict_reader.close()
        self.field_dict_reader.close()

class Index:

    "An inverted index solution encapsulating the various components."

    def __init__(self, pathname):
        self.pathname = pathname
        self.reader = None
        self.writer = None

    def get_writer(self, interval=TERM_INTERVAL, doc_interval=DOCUMENT_INTERVAL, flush_interval=FLUSH_INTERVAL):

        """
        Return a writer, optionally using the given indexing 'interval',
        'doc_interval' and 'flush_interval'.
        """

        self._ensure_directory()
        self.writer = IndexWriter(self.pathname, interval, doc_interval, flush_interval)
        return self.writer

    def _ensure_directory(self):
        if not exists(self.pathname):
            mkdir(self.pathname)

    def get_reader(self, partition=0):

        "Return a reader for the index."

        # Ensure that only one partition exists.

        self.merge()
        return self._get_reader(partition)

    def _get_reader(self, partition):

        "Return a reader for the index."

        if not exists(self.pathname):
            raise OSError, "Index path %r does not exist." % self.pathname

        self.reader = IndexReader(self.pathname)
        return self.reader

    def _get_partitions(self, prefix):

        """
        Return a set of partition identifiers using 'prefix' to identify
        relevant files.
        """

        prefix_length = len(prefix)

        partitions = set()
        for filename in listdir(self.pathname):
            if filename.startswith(prefix):
                partitions.add(filename[prefix_length:])
        return partitions

    def get_term_partitions(self):

        "Return a set of term partition identifiers."

        return self._get_partitions("terms-")

    def get_field_partitions(self):

        "Return a set of field partition identifiers."

        return self._get_partitions("fields-")

    def merge(self):

        "Merge/optimise index partitions."

        self._merge_terms()
        self._merge_fields()

    def _merge_terms(self, interval=TERM_INTERVAL, doc_interval=DOCUMENT_INTERVAL):

        """
        Merge term dictionaries using the given indexing 'interval' and
        'doc_interval'.
        """

        readers = []
        partitions = self.get_term_partitions()

        for partition in partitions:
            readers.append(get_term_reader(self.pathname, partition))

        # Write directly to a dictionary.

        if len(readers) > 1:
            if "merged" in partitions:
                rename_term_files(self.pathname, "merged", "old-merged")
                partitions.remove("merged")
                partitions.add("old-merged")

            writer = get_term_writer(self.pathname, "merged", interval, doc_interval)
            merger = TermDictionaryMerger(writer, readers)
            merger.merge()
            merger.close()

            # Remove old files.

            for partition in partitions:
                remove_term_files(self.pathname, partition)

        elif len(readers) == 1:
            partition = list(partitions)[0]
            if partition != "merged":
                rename_term_files(self.pathname, partition, "merged")

    def _merge_fields(self, interval=FIELD_INTERVAL):

        "Merge field dictionaries using the given indexing 'interval'."

        readers = []
        partitions = self.get_field_partitions()

        for partition in partitions:
            readers.append(get_field_reader(self.pathname, partition))

        # Write directly to a dictionary.

        if len(readers) > 1:
            if "merged" in partitions:
                rename_field_files(self.pathname, "merged", "old-merged")
                partitions.remove("merged")
                partitions.add("old-merged")

            writer = get_field_writer(self.pathname, "merged", interval)
            merger = FieldDictionaryMerger(writer, readers)
            merger.merge()
            merger.close()

            # Remove old files.

            for partition in partitions:
                remove_field_files(self.pathname, partition)

        elif len(readers) == 1:
            partition = list(partitions)[0]
            if partition != "merged":
                rename_field_files(self.pathname, partition, "merged")

    def update(self, other_indexes):

        "Copy the content of the 'other_indexes' into this index and merge."

        self._ensure_directory()

        for i, index in enumerate(other_indexes):
            for partition in index.get_term_partitions():
                copy_term_files(index.pathname, partition, self.pathname, "-added-%d" % i)
            for partition in index.get_field_partitions():
                copy_field_files(index.pathname, partition, self.pathname, "-added-%d" % i)

        self.merge()

    def close(self):
        if self.reader is not None:
            self.reader.close()
            self.reader = None
        if self.writer is not None:
            self.writer.close()
            self.writer = None

# vim: tabstop=4 expandtab shiftwidth=4
